# import numpy as np # type: ignore
# import attr
# from typing import Any, Optional, Sequence, Sized, Union, Iterable
# # import scipy
# # import scipy.optimize
# # from numpy.linalg import norm
# # from collections import namedtuple
# # from mdh.link import mdh_params, RevoluteLink, JointType
# # from scipy.spatial.transform import Rotation as R
# # from math import atan2, pi
# from mdh.kinematic_chain import KinematicChain
#
# @attr.s
# class Leg(KinematicChain):
#     """
#     inherets fk/ik, read/write json
#     has:
#         - foot position, orientation [m, quaternion???]
#         - joint angles [rads]
#     """
#     foot = attr.ib(type=Sequence[float])
#
#     # @property
#     # def



import attr
from collections import namedtuple
from enum import IntFlag
import numpy as np # type: ignore

JointType =  IntFlag('JointType', 'revolute prismatic revolute_theda revolute_alpha')
mdh_params = namedtuple("mdh_params", "alpha a theta d type")

@attr.s
class Link:
    """
    why? what is the value of this?
    prismatic/revolute have parameters that never chance ... need to freeze them!
    """
    # alpha = attr.ib(type=float)
    # a = attr.ib(type=float)
    # theta = attr.ib(type=float)
    # d = attr.ib(type=float)

@attr.s
class RevoluteLink(Link):
    """
    RevoluteLink about theta. All other parameters (alpha, a, d) are fixed and
    cannot be changed once the link is created.
    """
    _alpha = attr.ib(type=float)
    _a = attr.ib(type=float)
    theta = attr.ib(type=float)
    _d = attr.ib(type=float)

    @property
    def alpha(self) -> float:
        return self._alpha

    @property
    def a(self) -> float:
        return self._a

    @property
    def d(self) -> float:
        return self._d

    # max/min limits of link for optimzation
    min = attr.ib(-np.pi, type=float)
    max = attr.ib(np.pi, type=float)

    """Some of these params are immutable, how do I do that?"""
    def transform(self, angle: float):
        """could move this to link"""

        crx = np.cos(self._alpha)
        srx = np.sin(self._alpha)
        crz = np.cos(angle)
        srz = np.sin(angle)

        d = self._d
        a = self._a
        # self.theda = angle

        transform = np.array(
            [
                [crz, -srz, 0, a],
                [crx * srz, crx * crz, -srx, -d * srx],
                [srx * srz, crz * srx, crx, d * crx],
                [0, 0, 0, 1],
            ],
            dtype=np.float64,
        )

        return transform

@attr.s
class Prismatic(Link):
    min = attr.ib(type=float)
    max = attr.ib(type=float)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mdh']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=19.0', 'numpy>=1.9', 'scipy>=1.0']

setup_kwargs = {
    'name': 'mdh',
    'version': '0.1.1',
    'description': 'modified dh',
    'long_description': '# Modified Denavit–Hartenberg (mdh)\n\n[![Actions Status](https://github.com/MultipedRobotics/dh/workflows/CheckPackage/badge.svg)](https://github.com/MultipedRobotics/dh/actions)\n![GitHub](https://img.shields.io/github/license/multipedrobotics/dh)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/mdh)\n![PyPI](https://img.shields.io/pypi/v/mdh)\n\n\n<img src="https://upload.wikimedia.org/wikipedia/commons/d/d8/DHParameter.png" width="600px">\n\n[Modified Denavit-Hartenberg parameters](https://en.wikipedia.org/wiki/Denavit%E2%80%93Hartenberg_parameters#Modified_DH_parameters)\n\n## Inspiration\n\nYou should probably use one of these, they inspired me to write a simpler\nmodule for my needs:\n\n- [pybotics](https://github.com/nnadeau/pybotics)\n- [pytransform3d](https://github.com/rock-learning/pytransform3d)\n',
    'author': 'walchko',
    'author_email': 'walchko@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://pypi.org/project/mdh/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6',
}


setup(**setup_kwargs)



import numpy as np # type: ignore
import attr
from typing import Any, Optional, Sequence, Sized, Union, Iterable, Dict, List
import scipy # type: ignore
import scipy.optimize # type: ignore
# from numpy.linalg import norm
from collections import namedtuple
from mdh.link import mdh_params, RevoluteLink, JointType
from scipy.spatial.transform import Rotation as R # type: ignore
from math import atan2, pi

"""

"""

# Pose = namedtuple('Pose', "position rotation")

@attr.s
class KinematicChain(Sized, Iterable):
    _links = attr.ib(type=Sequence[Any])

    def __len__(self):
        return len(self._links)

    def __iter__(self):
        for l in self._links:
            yield l

    def transform(self, joints: Sequence[float]):
        """Calculates the transformation and returns a 4x4 matrix"""
        if len(joints) != len(self._links):
            raise Exception("inputs don't equal number of links")

        t = np.eye(4)
        for l, q in zip(reversed(self._links), reversed(joints)):
            m = l.transform(q)
            t = m.dot(t)
            # t = m @ t
            # t = np.dot(m, t)
        return t

    def forward(self, joints: Sequence[float]):
        """value?
        Solve the forward kinematics and returns a 4x4 matrix
        """
        return self.transform(joints)

    def inverse(self, pos: Sequence[float]) -> Optional[np.ndarray]:
            """
            Solve the inverse kinematics and returns 1xN on success or None
            on failure.

            FIXME: this only works for when the foot is straight down
            """
            j = [[],[]] # type: Sequence[List]
            for l in self:
                j[0].append(l.min)
                j[1].append(l.max)
            j[0][4] = -0.00001
            j[1][4] = 0.00001

            joint_limits = np.array(j) # type: np.ndarray
            # print(joint_limits)

            # r = R.from_matrix([[1,0,0],[0,0,-1],[0,1,0]])
            # rr = R.from_euler('z', atan2(pos[1], pos[0]))
            # c = rr*r

            # make desired transform pose (position, oritentation)
            # change the -90 x-axis to something else for non-down??
            x = pos[0]
            y = pos[1]
            c = R.from_euler('zyx', [-90,atan2(y,x)*180/pi,90], degrees=True)
            ppos = np.eye(4)
            ppos[:3,:3] = c.as_matrix()
            ppos[:3, 3] = pos

            start = [0]*len(self._links)

            result = scipy.optimize.least_squares(
                fun=_ik_cost_function,
                x0=np.array(start),
                bounds=joint_limits,
                args=(ppos, self)
            )  # type: scipy.optimize.OptimizeResult

            if result.success:  # pragma: no cover
                # print(">> yeah!!!")
                # print(">>", np.rad2deg(result.x))
                actual_t = self.transform(result.x)
                actual_pos = actual_t[:3,3]
                if np.allclose(actual_pos, pos, atol=1e-3):
                    return result.x
            return None

    @classmethod
    def from_parameters(cls: Any, params: Union[Sequence[Dict], Sequence[mdh_params]]):
        """Builds a KinematicChain object from an input"""
        links = []
        for l in params:
            if isinstance(l, dict):
                l = mdh_params(l['alpha'], l['a'], l['theta'], l['d'], l['type'])
            # print(f">> {l}")
            if l.type == JointType.revolute:
                link = RevoluteLink(a=l.a, alpha=l.alpha, d=l.d, theta=l.theta)
                # print(link.transform(0),'\n')
                links.append(link)
            elif l.type == JointType.prismatic:
                raise NotImplementedError(f"from_parameters: {l.type}")
            else:
                raise Exception(f"Invalid parameter: {l.type}")

        ret = cls(links)

        return ret

def _ik_cost_function(ik_q: np.ndarray, pose: np.ndarray, kc: KinematicChain) -> np.ndarray:
    """Cost function for least squares algorithm.
        - ik_q: calculated 4x4 matrix from optimization
        - pose: desired 4x4 matrix with position and orientation
        - kc: KinematicChain object that is trying to find this info
    """
    actual_t = kc.transform(ik_q)
    diff = np.abs(actual_t - pose)
    return diff.ravel()
